from fares.dimension.dimension import SlotDimensions, UserDimensions, ClusterDimensions


def inter_cluster_resolver(slot):
    return {
        "slot.od_cluster": (
            slot[SlotDimensions.ORIGIN_CLUSTER_ID],
            slot[SlotDimensions.DESTINATION_CLUSTER_ID],
        ),
        "slot.origin_cluster_id": slot[SlotDimensions.ORIGIN_CLUSTER_ID],
        "slot.destination_cluster_id": slot[SlotDimensions.DESTINATION_CLUSTER_ID],
        "slot.route_id": slot[SlotDimensions.ROUTE_ID],
        "slot.start_time": slot[SlotDimensions.START_TIME],
    }


def zone_resolver(slot):
    return {
        "slot.route_id": slot[SlotDimensions.ROUTE_ID],
        "slot.start_time": slot[SlotDimensions.START_TIME],
        "slot.zone_id": slot.get(SlotDimensions.ZONE_ID),
    }


def resolve_user_dimensions(user):
    return {
        "user.blacklisted": user.get(UserDimensions.BLACKLISTED, False),
        "user.pass_purchase_count": user[UserDimensions.PASS_PURCHASE_COUNT],
        "user.id": user[UserDimensions.ID],
        "user.new_device": user[UserDimensions.NEW_DEVICE],
    }


def resolve_cluster_dimensions(clusters):
    return {
        "clusters.od_cluster": (
            clusters[ClusterDimensions.ORIGIN_CLUSTER_ID],
            clusters[ClusterDimensions.DESTINATION_CLUSTER_ID],
        ),
        "clusters.zone_id": clusters[ClusterDimensions.ZONE_ID],
    }



from dataclasses import dataclass
from datetime import datetime
from enum import Enum

from rule_engine.operator import In, Equals, AnyIn
from rule_engine.rule import RuleSet, Rule
from typing import List, Union


@dataclass(frozen=True)
class InterClusterApplicability:
    origin_cluster_id: str
    destination_cluster_id: str


@dataclass
class ZoneWideApplicability:
    zone_id: str


@dataclass
class Applicability:
    type: str
    rules: List[Union[InterClusterApplicability, ZoneWideApplicability]]


class SubscriptionType(Enum):
    TRIAL = "TRIAL"
    STANDARD = "STANDARD"
    HIDDEN = "HIDDEN"


@dataclass
class SubscriptionOffering:
    id: str
    name: str
    rides: int
    currency: str
    validity_in_days: int
    is_carry_forward: bool
    activation_date: datetime
    deactivation_date: datetime
    zone_id: str
    amount: int
    subscription_type: SubscriptionType
    applicability: Applicability
    created_at: datetime

    @property
    def applicability_type(self):
        return self.applicability.type

    @property
    def applicability_rules(self):
        return self.applicability.rules

    def usage_applicability_rules(self) -> RuleSet:
        rules = []
        if self.applicability_type == "ZONE_WIDE":
            rule = Rule(
                "slot.zone_id", In(), [a.zone_id for a in self.applicability_rules]
            )
            rules.append(rule)
        elif self.applicability_type == "INTER_CLUSTER":
            origin_clusters = {
                app.origin_cluster_id for app in self.applicability_rules
            }
            destination_clusters = {
                app.destination_cluster_id for app in self.applicability_rules
            }
            all_clusters = origin_clusters.union(destination_clusters)
            # Business rule currently states that the user can travel if booking origin or destination cluster,
            # matches the subscription origin or destination cluster
            od_match_rule = Rule("slot.od_cluster", AnyIn(), all_clusters)
            rules.append(od_match_rule)

        return RuleSet(rules)

    def purchase_applicability_rules(self) -> RuleSet:
        rules = []

        if SubscriptionType.TRIAL == self.subscription_type:
            rules.append(Rule("user.blacklisted", Equals(), False))
            rules.append(Rule("user.pass_purchase_count", Equals(), 0))
            rules.append(Rule("user.new_device", Equals(), True))

        return RuleSet(rules)

    def cluster_applicability_rules(self) -> RuleSet:
        rules = []
        if self.applicability_type == "ZONE_WIDE":
            rule = Rule(
                "clusters.zone_id", In(), [a.zone_id for a in self.applicability_rules]
            )
            rules.append(rule)
        elif self.applicability_type == "INTER_CLUSTER":
            od_list = [
                (app.origin_cluster_id, app.destination_cluster_id)
                for app in self.applicability_rules
            ]
            reverse_od_list = [
                (app.destination_cluster_id, app.origin_cluster_id)
                for app in self.applicability_rules
            ]
            valid_od_combinations = od_list + reverse_od_list
            # Business use case currently states that if a user can go from cluster A to B using a subscription, the
            # user shd be able to travel from cluster B to A
            od_match_rule = Rule("clusters.od_cluster", In(), valid_od_combinations)
            rules.append(od_match_rule)

        return RuleSet(rules)

    @classmethod
    def from_json(cls, dikt):
        def _inter_cluster(rules):
            applicability_rules = {
                (app["origin_cluster_id"], app["destination_cluster_id"])
                for app in rules
            }
            return [
                InterClusterApplicability(applicability_rule[0], applicability_rule[1])
                for applicability_rule in applicability_rules
            ]

        def _zone_wide(rules):
            return [ZoneWideApplicability(rule["zone_id"]) for rule in rules]

        def _applicability(dikt):
            type = dikt["type"]
            rules = dikt["rules"]
            if type == "ZONE_WIDE":
                return Applicability("ZONE_WIDE", _zone_wide(rules))
            elif type == "INTER_CLUSTER":
                return Applicability("INTER_CLUSTER", _inter_cluster(rules))

        return SubscriptionOffering(
            id=dikt["id"],
            name=dikt["name"],
            rides=dikt["rides"],
            currency=dikt["currency"],
            validity_in_days=dikt["validity_in_days"],
            is_carry_forward=dikt["is_carry_forward"],
            activation_date=dikt["activation_date"],
            deactivation_date=dikt["deactivation_date"],
            zone_id=dikt["zone_id"],
            amount=dikt["amount"],
            subscription_type=SubscriptionType(dikt["subscription_type"]),
            applicability=_applicability(dikt["applicability"]),
            created_at=datetime.fromisoformat(dikt["created_at"]),
        )



import pytest

from fares import SubscriptionOffering, is_subscription_offering_applicable_for_clusters
from fares.dimension import ClusterDimensions


@pytest.fixture
def subscription_offering():
    return {
        "id": "first_sub",
        "name": "Morning sub",
        "rides": "10",
        "currency": "INR",
        "validity_in_days": "20",
        "is_carry_forward": "True",
        "activation_date": "2019-02-28T12:56:41.589953+00:00",
        "deactivation_date": "2019-02-28T12:56:41.589953+00:00",
        "zone_id": "zone1",
        "amount": "1000",
        "subscription_type": "STANDARD",
        "applicability": {
            "type": "INTER_CLUSTER",
            "rules": [
                {
                    "origin_cluster_id": "cluster_1",
                    "destination_cluster_id": "cluster_2",
                    "session": "MORNING",
                }
            ],
        },
        "created_at": "2019-02-28T12:56:41.589953+00:00",
    }


@pytest.fixture
def subscription_offering_zone_applicability():
    return {
        "id": "first_sub",
        "name": "Morning sub",
        "rides": "10",
        "currency": "INR",
        "validity_in_days": "20",
        "is_carry_forward": "True",
        "activation_date": "2019-02-28T12:56:41.589953+00:00",
        "deactivation_date": "2019-02-28T12:56:41.589953+00:00",
        "zone_id": "zone-1",
        "amount": "1000",
        "subscription_type": "TRIAL",
        "applicability": {"type": "ZONE_WIDE", "rules": [{"zone_id": "zone_1"}]},
        "created_at": "2019-02-28T12:56:41.589953+00:00",
    }


@pytest.fixture
def valid_clusters():
    return {
        ClusterDimensions.ORIGIN_CLUSTER_ID: "cluster_1",
        ClusterDimensions.DESTINATION_CLUSTER_ID: "cluster_2",
        ClusterDimensions.ZONE_ID: "zone_1",
    }


@pytest.fixture
def valid_reverse_clusters():
    return {
        ClusterDimensions.ORIGIN_CLUSTER_ID: "cluster_2",
        ClusterDimensions.DESTINATION_CLUSTER_ID: "cluster_1",
        ClusterDimensions.ZONE_ID: "zone_1",
    }


@pytest.fixture
def invalid_clusters():
    return {
        ClusterDimensions.ORIGIN_CLUSTER_ID: "cluster_1",
        ClusterDimensions.DESTINATION_CLUSTER_ID: "cluster_3",
        ClusterDimensions.ZONE_ID: "zone_1",
    }


@pytest.fixture
def invalid_zone_clusters():
    return {
        ClusterDimensions.ORIGIN_CLUSTER_ID: "cluster_1",
        ClusterDimensions.DESTINATION_CLUSTER_ID: "cluster_3",
        ClusterDimensions.ZONE_ID: "zone_2",
    }


def test_cluster_are_applicable(subscription_offering, valid_clusters):
    assert is_subscription_offering_applicable_for_clusters(
        SubscriptionOffering.from_json(subscription_offering), valid_clusters
    )


def test_reverse_clusters_are_applicable(subscription_offering, valid_reverse_clusters):
    assert is_subscription_offering_applicable_for_clusters(
        SubscriptionOffering.from_json(subscription_offering), valid_reverse_clusters
    )


def test_invalid_clusters_are_not_applicable(subscription_offering, invalid_clusters):
    assert not is_subscription_offering_applicable_for_clusters(
        SubscriptionOffering.from_json(subscription_offering), invalid_clusters
    )


def test_zone_applicability(subscription_offering_zone_applicability, valid_clusters):
    assert is_subscription_offering_applicable_for_clusters(
        SubscriptionOffering.from_json(subscription_offering_zone_applicability),
        valid_clusters,
    )


def test_zone_applicability_fails(
    subscription_offering_zone_applicability, invalid_zone_clusters
):
    assert not is_subscription_offering_applicable_for_clusters(
        SubscriptionOffering.from_json(subscription_offering_zone_applicability),
        invalid_zone_clusters,
    )
